@@ -152,7 +152,7 @@ color_swatch_size_allocate (GtkWidget *widget,
 
 	priv = GTKHTML_COLOR_SWATCH_GET_PRIVATE (widget);
 
-	widget->allocation = *allocation;
+	gtk_widget_set_allocation (widget, allocation);
 	gtk_widget_size_allocate (priv->frame, allocation);
 }
 
@@ -179,8 +179,8 @@ color_swatch_class_init (GtkhtmlColorSwatchClass *class)
 		PROP_COLOR,
 		g_param_spec_boxed (
 			"color",
-			_("Color"),
-			_("The current color"),
+			"Color",
+			"The current color",
 			GDK_TYPE_COLOR,
 			G_PARAM_READWRITE));
 
@@ -189,8 +189,8 @@ color_swatch_class_init (GtkhtmlColorSwatchClass *class)
 		PROP_SHADOW_TYPE,
 		g_param_spec_enum (
 			"shadow-type",
-			_("Frame Shadow"),
-			_("Appearance of the frame border"),
+			"Frame Shadow",
+			"Appearance of the frame border",
 			GTK_TYPE_SHADOW_TYPE,
 			GTK_SHADOW_ETCHED_IN,
 			G_PARAM_READWRITE));
