@@ -461,7 +461,7 @@ delete_setup_undo (HTMLEngine *e, HTMLObject *buffer, guint len, guint position_
 #ifdef OP_DEBUG
 	printf ("delete undo len %d\n", len);
 #endif
-	html_undo_add_action (e->undo,
+	html_undo_add_action (e->undo, e,
 			      html_undo_action_new ("Delete object", delete_undo_action,
 						    HTML_UNDO_DATA (undo), html_cursor_get_position (e->cursor),
 						    position_after),
@@ -721,7 +721,7 @@ fix_empty_aligned_setup_undo (HTMLEngine *e, HTMLUndoDirection dir, HTMLObject *
 	undo->data.destroy = fix_empty_aligned_undo_destroy;
 	undo->ac       = ac;
 
-	html_undo_add_action (e->undo,
+	html_undo_add_action (e->undo, e,
 			      html_undo_action_new ("Remove empty aligned", fix_empty_aligned_undo_action,
 						    HTML_UNDO_DATA (undo), html_cursor_get_position (e->cursor),
 						    html_cursor_get_position (e->cursor)),
@@ -817,7 +817,7 @@ html_engine_cut (HTMLEngine *e)
 		gint start_position = start->position;
 		gint end_position = end->position;
 		if (end_position - start_position > 0) {
-			int len = end_position - start_position;
+			gint len = end_position - start_position;
 			g_signal_emit_by_name (e->widget, "object_delete", start_position, len);
 		}
 		html_cursor_destroy (start);
@@ -825,7 +825,7 @@ html_engine_cut (HTMLEngine *e)
 	}
 
 	rv = delete_object (e, &e->clipboard, &e->clipboard_len, HTML_UNDO_UNDO, TRUE);
-	html_undo_level_end (e->undo);
+	html_undo_level_end (e->undo, e);
 
 #ifdef OP_DEBUG
 	printf ("cut  len: %d\n", e->clipboard_len);
@@ -989,7 +989,7 @@ insert_setup_undo (HTMLEngine *e, guint len, guint position_before, HTMLUndoDire
 
 	/* printf ("insert undo len %d\n", len); */
 
-	html_undo_add_action (e->undo,
+	html_undo_add_action (e->undo, e,
 			      html_undo_action_new ("Insert", insert_undo_action,
 						    HTML_UNDO_DATA (undo),
 						    html_cursor_get_position (e->cursor),
@@ -1024,7 +1024,7 @@ fix_aligned_undo_action (HTMLEngine *e, HTMLUndoData *data, HTMLUndoDirection di
 	html_clue_remove (HTML_CLUE (cf->parent), cf);
 	html_object_destroy (cf);
 
-	html_undo_add_action (e->undo,
+	html_undo_add_action (e->undo, e,
 			      html_undo_action_new ("Fix aligned", fix_aligned_redo_action,
 						    undo, html_cursor_get_position (e->cursor),
 						    position_before),
@@ -1041,7 +1041,7 @@ fix_align_setup_undo (HTMLEngine *e, guint position_before, HTMLUndoDirection di
 	html_undo_data_init (HTML_UNDO_DATA (undo));
 	/* printf ("insert undo len %d\n", len); */
 
-	html_undo_add_action (e->undo,
+	html_undo_add_action (e->undo, e,
 			      html_undo_action_new ("Undo aligned fix", fix_aligned_undo_action,
 						    undo, html_cursor_get_position (e->cursor),
 						    position_before),
@@ -1153,7 +1153,7 @@ html_engine_paste_object (HTMLEngine *e, HTMLObject *o, guint len)
 	html_undo_level_begin (e->undo, "Paste", "Paste");
 	html_engine_delete (e);
 	html_engine_insert_object (e, o, len, html_engine_get_insert_level_for_object (e, o));
-	html_undo_level_end (e->undo);
+	html_undo_level_end (e->undo, e);
 }
 
 void
@@ -1175,7 +1175,7 @@ check_magic_link (HTMLEngine *e, const gchar *text, guint len)
 	    && gtk_html_get_magic_links (e->widget)
 	    && len == 1
 	    && (*text == ' ' || text [0] == '\n' || text [0] == '>' || text [0] == ')'))
-		html_text_magic_link (HTML_TEXT (e->cursor->object), e, html_object_get_length (e->cursor->object));
+		html_text_magic_link (HTML_TEXT (e->cursor->object), e, 1);
 }
 
 static void
@@ -1242,7 +1242,7 @@ insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir, gboolean add_undo)
 
 	gtk_html_editor_event_command (e->widget, GTK_HTML_COMMAND_INSERT_PARAGRAPH, FALSE);
 	if (add_undo)
-		html_undo_level_end (e->undo);
+		html_undo_level_end (e->undo, e);
 
 	html_engine_thaw (e);
 	g_signal_emit_by_name (e->widget, "object_inserted", 0, 0);
@@ -1266,7 +1266,7 @@ html_engine_insert_empty_paragraph (HTMLEngine *e)
 	html_engine_thaw (e);
 }
 
-static char *picto_chars =
+static const gchar *picto_chars =
 	/*  0 */ "DO)(|/PQ*!"
 	/* 10 */ "S\0:-\0:\0:-\0"
 	/* 20 */ ":\0:;=-\"\0:;"
@@ -1284,7 +1284,7 @@ static gint picto_states [] = {
 	/* 50 */   0, -19,   0, -11,  56,   0, -11,   0, -13,  61,
 	/* 60 */   0, -13,   0,  -6,   0,  68,  -7,   0,  -7,   0,
 	/* 70 */ -16,  73,   0, -16,   0, -21,  78,   0, -21,   0 };
-static gchar *picto_icon_names [] = {
+static const gchar *picto_icon_names [] = {
 	"face-angel",
 	"face-angry",
 	"face-cool",
@@ -1445,7 +1445,7 @@ html_engine_insert_text_with_extra_attributes (HTMLEngine *e, const gchar *text,
 			text = nl + 1;
 		}
 	} while (nl);
-	html_undo_level_end (e->undo);
+	html_undo_level_end (e->undo, e);
 }
 
 void
@@ -1465,7 +1465,7 @@ html_engine_paste_text_with_extra_attributes (HTMLEngine *e, const gchar *text,
 	g_free (redo_name);
 	html_engine_delete (e);
 	html_engine_insert_text_with_extra_attributes (e, text, len, attrs);
-	html_undo_level_end (e->undo);
+	html_undo_level_end (e->undo, e);
 }
 
 void
@@ -1475,9 +1475,9 @@ html_engine_paste_text (HTMLEngine *e, const gchar *text, guint len)
 }
 
 void
-html_engine_paste_link (HTMLEngine *e, const char *text, int len, const char *complete_url)
+html_engine_paste_link (HTMLEngine *e, const gchar *text, gint len, const gchar *complete_url)
 {
-	char *url, *target;
+	gchar *url, *target;
 
 	if (len == -1)
 		len = g_utf8_strlen (text, -1);
@@ -1490,7 +1490,7 @@ html_engine_paste_link (HTMLEngine *e, const char *text, int len, const char *co
 	}
 
 	html_engine_paste_text (e, text, len);
-	html_text_add_link (HTML_TEXT (e->cursor->object), e, url, target, e->cursor->offset - len, e->cursor->offset);
+	html_text_add_link (HTML_TEXT (e->cursor->object), e, url, target, e->cursor->offset >= len ? e->cursor->offset - len : 0, e->cursor->offset);
 
 	g_free (url);
 }
@@ -1533,7 +1533,7 @@ html_engine_delete_n (HTMLEngine *e, guint len, gboolean forward)
 				html_cursor_backward (e->cursor, e);
 				html_engine_delete (e);
 				html_engine_insert_text (e, picto, -1);
-				html_undo_level_end (e->undo);
+				html_undo_level_end (e->undo, e);
 				g_free (picto);
 
 				html_engine_unblock_selection (e);
@@ -1542,7 +1542,7 @@ html_engine_delete_n (HTMLEngine *e, guint len, gboolean forward)
 			}
 		}
 		if (forward) {
-			int i;
+			gint i;
 
 			for (i = len; i > 0; i--)
 				html_cursor_forward (e->cursor, e);
@@ -1569,7 +1569,7 @@ html_engine_cut_line (HTMLEngine *e)
 		html_cursor_forward (e->cursor, e);
 
 	html_engine_cut (e);
-	html_undo_level_end (e->undo);
+	html_undo_level_end (e->undo, e);
 }
 
 typedef struct {
@@ -1684,7 +1684,7 @@ html_engine_append_object (HTMLEngine *e, HTMLObject *o, guint len)
 {
 	html_undo_level_begin (e->undo, "Append object", "Remove appended object");
 	append_object (e, o, len, HTML_UNDO_UNDO);
-	html_undo_level_end (e->undo);
+	html_undo_level_end (e->undo, e);
 }
 
 static void
@@ -1753,7 +1753,7 @@ html_engine_append_flow (HTMLEngine *e, HTMLObject *o, guint len)
 {
 	html_undo_level_begin (e->undo, "Append flow", "Remove appended flow");
 	append_flow (e, o, len, HTML_UNDO_UNDO);
-	html_undo_level_end (e->undo);
+	html_undo_level_end (e->undo, e);
 }
 
 void
@@ -1788,7 +1788,7 @@ html_engine_cut_and_paste_end (HTMLEngine *e)
 		insert_object (e, e->clipboard, e->clipboard_len, position, level, HTML_UNDO_UNDO, TRUE);
 		e->clipboard = NULL;
 	}
-	html_undo_level_end (e->undo);
+	html_undo_level_end (e->undo, e);
 	html_engine_clipboard_pop (e);
 	html_engine_selection_pop (e);
 	html_engine_show_cursor (e);
@@ -1865,7 +1865,7 @@ html_engine_delete (HTMLEngine *e)
 		gint end_position = end->position;
 
 		if (end_position - start_position > 0) {
-			int len = end_position - start_position;
+			gint len = end_position - start_position;
 			g_signal_emit_by_name (e->widget, "object_delete", start_position, len);
 		}
 
@@ -1884,9 +1884,9 @@ html_engine_delete (HTMLEngine *e)
 
 				/* go thru current cluev */
 				do {
-				        /* go thru current flow */
+					/* go thru current flow */
 					while (cur) {
-				                /* lets look if container is whole contained in the selection */
+						/* lets look if container is whole contained in the selection */
 						if (html_object_is_container (cur)) {
 							html_cursor_jump_to (e->cursor, e, cur, html_object_get_length (cur));
 							if (e->cursor->position > end->position) {
@@ -1903,7 +1903,7 @@ html_engine_delete (HTMLEngine *e)
 				} while (prev && prev->parent->next && (cur = html_object_head (prev->parent->next)));
 
 				if (prev) {
-				        /* cluev end is in the selection. Lets handle this case just like simple delete 
+					/* cluev end is in the selection. Lets handle this case just like simple delete
 					since text is the selection itself*/
 					if (e->mark)
 						html_cursor_destroy (e->mark);
@@ -1923,7 +1923,6 @@ html_engine_delete (HTMLEngine *e)
 			html_cursor_destroy (end);
 		html_cursor_jump_to_position (e->cursor, e, start_position);
 
-
 	}
-	html_undo_level_end (e->undo);
+	html_undo_level_end (e->undo, e);
 }
